import unittest
import tempfile
import shutil
from os import makedirs
from os.path import abspath, dirname, join

from kodi_addon_checker.handle_files import find_file
from kodi_addon_checker.handle_files import create_file_index
from kodi_addon_checker.handle_files import find_files_recursive as FFR

HERE = abspath(dirname(__file__))


class TestFindFile(unittest.TestCase):

    def __init__(self, methodName='runTest'):
        super().__init__(methodName)
        self.path = None
        self.file = None

    def setUp(self):
        self.path = tempfile.mkdtemp()
        self.file = join(self.path, "addon.txt")

    def test_find_file_is_none(self):
        self.assertIsNone(find_file("randomname", self.path))

    def test_find_file_is_not_none(self):
        with open(self.file, "w+", encoding="utf8"):
            self.assertIsNotNone(find_file("addon.txt", self.path))

    def teardown(self):
        shutil.rmtree(self.path)


class TestFindFilesRecursive(unittest.TestCase):

    def __init__(self, methodName='runTest'):
        super().__init__(methodName)
        self.path = None
        self.file = None
        self.folder = None

    def setUp(self):
        self.path = tempfile.mkdtemp()
        self.file = join(self.path, "addon.txt")
        dirs = join(self.path, "tree/subdir")
        makedirs(dirs)
        self.folder = join(dirs, "addon.xml")

    def test_find_files_recursive_is_False(self):
        self.assertFalse(list(FFR("randomname", self.path)))

    def test_find_files_recursive_is_True(self):
        with open(self.file, "w+", encoding="utf8"):
            self.assertTrue(FFR("addon.txt", self.path))

    def test_find_files_recursive_deeply(self):
        with open(self.folder, "w+", encoding="utf8"):
            self.assertTrue(FFR("addon.xml", self.path))

    def teardown(self):
        shutil.rmtree(self.path)
        shutil.rmtree(self.folder)


class TestCreateFileIndex(unittest.TestCase):

    def test_create_file_index(self):
        path = join(HERE, 'test_data', 'File_index/')
        expected_list = [{'path': path, 'name': 'file_index.py'}]
        output = create_file_index(path)
        self.assertListEqual(output, expected_list)



# -*- coding: utf-8 -*-
#

from kodi_addon_checker.check_allowed_versions import version_is_valid

def test_addonversion_simple():
    assert version_is_valid("1.0.1")


def test_addonversion_localversionidentifier():
    assert version_is_valid("1.0.1+matrix.1")


def test_addonversion_localversionidentifier2():
    assert version_is_valid("1.0.1+matrix.2")


def test_invalidversion():
    assert not version_is_valid("someinvalidversion")


def test_invalidbackportversion():
    assert not version_is_valid("2.3.0-backported-Leia")



# -*- coding: utf-8 -*-
#

import os

from PIL import Image

from kodi_addon_checker.common import has_transparency
from kodi_addon_checker.config import Config

FIXTURE_PATH = os.path.join("tests", "fixtures")


def __load_image(filename):
    filepath = os.path.join(FIXTURE_PATH, filename)
    return Image.open(filepath)


def test_if_false_gets_picked_up():
    config = Config(FIXTURE_PATH)
    assert config.is_enabled("check_license_file_exists") is False


def test_if_true_gets_picked_up():
    config = Config(FIXTURE_PATH)
    assert config.is_enabled("true_config") is True


def test_if_does_not_exists_default_to_false():
    config = Config(FIXTURE_PATH)
    assert config.is_enabled("does_not_exist") is False


def test_with_missing_config():
    config = Config('does_not_exist')
    assert config.is_enabled("does_not_exist") is False


def test_has_transparency_rgb():
    image = __load_image("rgb_icon.png")
    assert has_transparency(image) is False


def test_has_transparency_rgba():
    image = __load_image("rgba_icon.png")
    assert has_transparency(image) is False


def test_has_transparency_rgba_transparency():
    image = __load_image("rgba_icon_transparency.png")
    assert has_transparency(image) is True



# -*- coding: utf-8 -*-
#

from kodi_addon_checker.versions import AddonVersion, KodiVersion


def test_addonversion_simple_patch():
    assert AddonVersion("1.0.1") > AddonVersion("1.0.0")


def test_addonversion_simple_minor():
    assert AddonVersion("1.1.0") > AddonVersion("1.0.0")


def test_addonversion_simple_major():
    assert AddonVersion("2.0.0") > AddonVersion("1.0.0")


def test_addonversion_kodi_versions():
    assert AddonVersion("1.0.0+matrix.1") > AddonVersion("1.0.0+leia.1")
    assert AddonVersion("1.0.0+leia.1") == AddonVersion("1.0.0+Leia.1")


def test_addonversion_kodi_simplevsstring():
    assert AddonVersion("0.5.4.1") > AddonVersion("0.5.4+matrix.1")


def test_kodi_non_compliant_pep440_builtin_versions():
    # mimics the tests in TestAddonVersion.cpp (non PEP440 compliant)
    # try to use PEP440 compliant versions whenever possible
    assert AddonVersion("1.0.0-1") > AddonVersion("1.0.0")
    assert AddonVersion("1.0.0~beta2") > AddonVersion("1.0.0~beta1")
    assert AddonVersion("1.0.1~beta01") > AddonVersion("1.0.0")
    assert AddonVersion("1.0.1~beta") > AddonVersion("1.0.0")
    assert AddonVersion("1.0.0~beta") > AddonVersion("1.0.0~alpha")
    assert AddonVersion("1.0.0~beta2") > AddonVersion("1.0.0~alpha")
    assert AddonVersion("1.0.0~alpha3") > AddonVersion("1.0.0~alpha2")


def test_kodiversions():
    assert KodiVersion("matrix") > KodiVersion("leia")



"""
    Copyright (C) 2017-2018 Team Kodi
    This file is part of Kodi - kodi.tv

    SPDX-License-Identifier: GPL-3.0-only
    See LICENSES/README.md for more information.
"""

import argparse
import logging
import os
import sys

from kodi_addon_checker import __version__, check_addon, ValidKodiVersions
from kodi_addon_checker.check_repo import check_repo
from kodi_addon_checker.common import load_plugins
from kodi_addon_checker.config import Config, ConfigManager
from kodi_addon_checker.logger import Logger
from kodi_addon_checker.record import INFORMATION, PROBLEM, WARNING, Record
from kodi_addon_checker.report import Report


def dir_type(dir_path):
    """ArgParse callable to validate positional add-on arguments

    Arguments:
        dir_path {string} -- user defined argument

    Raises:
        argparse.ArgumentTypeError -- thrown when user input is not a directory or does not contain addon.xml

    Returns:
        string -- absolute path of the input directory
    """
    if not os.path.isdir(dir_path):
        raise argparse.ArgumentTypeError(
            f"Add-on directory {dir_path} does not exist")
    return os.path.abspath(dir_path)


def check_artifact(artifact_path, args, all_repo_addons):
    """
    Check given artifact and return its report. The artifact can be either an add-on or a repository.
    :param artifact_path: the path of add-on or repo
    :param args: argparse object
    :return: report
    """
    logger = logging.getLogger(__package__)
    logger.info("Downloading all repo addon list")
    logger.info("Download completed")
    artifact_path = os.path.abspath(artifact_path)
    config = Config(artifact_path, args)
    ConfigManager.process_config(config)
    if os.path.isfile(os.path.join(artifact_path, "addon.xml")):
        return check_addon.start(artifact_path, args, all_repo_addons, config)

    return check_repo(artifact_path, args, all_repo_addons, config)


def main():
    """The entry point to kodi-addon-checker
    """
    load_plugins()
    parser = argparse.ArgumentParser(prog="kodi-addon-checker",
                                     description="Checks Kodi repo for best practices and creates \
                                     problem and warning reports.\r\nIf optional add-on \
                                     directories are provided, check only those add-ons. \
                                     Otherwise, scan current repository and check all add-ons in \
                                     the current directory.")
    parser.add_argument("--version", action="version",
                        version=f"%(prog)s {__version__}")
    parser.add_argument("dir", type=dir_type, nargs="*", help="optional add-on or repo directories")
    parser.add_argument("--branch", choices=ValidKodiVersions, required=True,
                        help="Target branch name where the checker will resolve dependencies")
    parser.add_argument("--PR", help="Tell if tool is to run on a pull requests or not", action='store_true')
    parser.add_argument("--allow-folder-id-mismatch", help="Allow the addon's folder name and id to mismatch",
                        action="store_true")
    parser.add_argument("--enable-debug-log", help="Enable debug logging to kodi-addon-checker.log",
                        action="store_true", default=False)
    parser.add_argument("--skip-dependency-checks", help="Do not check if addon dependencies are available \
                        in the official Kodi addon repository", action="store_true", default=False)
    ConfigManager.fill_cmd_args(parser)
    args = parser.parse_args()

    log_file_name = os.path.join(os.getcwd(), "kodi-addon-checker.log")
    Logger.create_logger(log_file_name, __package__, args.enable_debug_log)

    all_repo_addons = check_addon.get_all_repo_addons()

    if args.dir:
        # Following report is a wrapper for all sub reports
        report = Report("")
        for directory in args.dir:
            report.add(check_artifact(directory, args, all_repo_addons))
    else:
        report = check_artifact(os.getcwd(), args, all_repo_addons)

    if report.problem_count > 0:
        report.add(Record(PROBLEM, f"We found {report.problem_count} problems " \
            f"and {report.warning_count} warnings, please check the logfile."))
        sys.exit(1)
    elif report.warning_count > 0:
        report.add(Record(WARNING, "We found no problems "\
            f"and {report.warning_count} warnings, please check the logfile."))
    else:
        report.add(Record(INFORMATION, "We found no problems and no warnings, please enjoy your day."))


if __name__ == "__main__":
    main()



"""
    Copyright (C) 2017-2018 Team Kodi
    This file is part of Kodi - kodi.tv

    SPDX-License-Identifier: GPL-3.0-only
    See LICENSES/README.md for more information.
"""

from kodi_addon_checker.report import Record
from kodi_addon_checker.reporter import Reporter, reporter


@reporter(name="array", enabled=False)
class ArrayReporter(Reporter):
    def __init__(self):
        self.reports = []

    def report(self, report):
        if isinstance(report, Record):
            self.reports.append(report)



"""
    Copyright (C) 2017-2018 Team Kodi
    This file is part of Kodi - kodi.tv

    SPDX-License-Identifier: GPL-3.0-only
    See LICENSES/README.md for more information.
"""

import inspect
from abc import ABC


class Reporter(ABC):

    def report(self, report):
        pass


class ReportManager():
    reporters = {}

    @classmethod
    def register(cls, reporter_clazz: Reporter, name, enabled):
        cls.reporters[name] = [reporter_clazz(), enabled]

    @classmethod
    def enable(cls, names):
        """
        Enable only the given list of names and disable the rest.
        :param names: list of reporter names
        :return: None
        """
        for name, arr in cls.reporters.items():
            arr[1] = name in names

    @classmethod
    def getEnabledReporters(cls):
        return [reporter[0] for reporter in cls.reporters.values() if reporter[1]]


def reporter(name, enabled=False):
    def _reporter(clazz):
        if inspect.isclass(clazz):
            if not hasattr(clazz, "report") or len(inspect.signature(getattr(clazz, "report")).parameters.items()) != 2:
                raise RuntimeError("Reporter must have a function 'report(self, report: Report)")
        else:
            raise RuntimeError("Reporter must be a class")

        # Register the reporter
        ReportManager.register(clazz, name, enabled)
        return clazz

    return _reporter



import os
import setuptools

import kodi_addon_checker


with open("requirements.txt", "r", encoding="utf8") as f:
    requirements = f.read().splitlines()

_ROOT = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(_ROOT, "README.md"), "r", encoding="utf8") as f:
    LONG_DESCRIPTION = f.read()

setuptools.setup(
    name="kodi-addon-checker",
    version=kodi_addon_checker.__version__,
    description="Check kodi addons or whole kodi repositories for errors and best practices.",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author="Team Kodi",
    url="https://github.com/xbmc/addon-check",
    download_url="https://github.com/xbmc/addon-check/archive/master.zip",
    packages=setuptools.find_packages(exclude=["script.test", "tests*"]),
    package_data={"kodi_addon_checker": ["xml_schema/*.xsd"]},
    install_requires=requirements,
    python_requires=">=3.5",
    setup_requires=["setuptools>=38.6.0"],
    entry_points={
        "console_scripts": ["kodi-addon-checker = kodi_addon_checker.__main__:main"]
    },
    keywords="kodi add-on add-on_checker",
    classifiers=[
        "Operating System :: POSIX :: Linux",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Topic :: Utilities",
    ]
    + [
        f"Programming Language :: Python :: {x}"
        for x in "3.8 3.9 3.10 3.11 3.12 3.13".split()
    ],
)



"""
    Copyright (C) 2017-2018 Team Kodi
    This file is part of Kodi - kodi.tv

    SPDX-License-Identifier: GPL-3.0-only
    See LICENSES/README.md for more information.
"""

import os

from radon.raw import analyze

from .record import PROBLEM, WARNING, Record
from .report import Report


def check_complex_addon_entrypoint(report: Report, addon_path: str, parsed_xml, max_entrypoint_count: int):
    """check the complexity of the addon.xml file by counting the number of lines in the file.
        :addon_path: path of the addon
        :parsed_xml: parsed addon.xml file
        :max_entrypoint_count: :max_entypoint_line_count: max value allowed in any entrypoint file
    """

    for i in parsed_xml.findall("extension"):
        library = i.get("library")

        if library:
            filepath = os.path.join(addon_path, library)

            if not os.path.isdir(filepath):
                ext = os.path.splitext(filepath)

                if os.path.exists(filepath):
                    if ext[1] == '.py':
                        _number_of_lines(report, filepath, library,
                                         max_entrypoint_count)
                else:
                    report.add(
                        Record(PROBLEM, f"{library} Entry point does not exists"))


def _number_of_lines(report: Report, filepath: str, library: str, max_entrypoint_count: int):
    """Returns the number of logical lines of code in a given python file
        :filepath: Path of the python file
        :library: relative path of the file
        :max_entypoint_line_count: max value allowed in any entrypoint file
    """

    try:
        with open(filepath, 'r', encoding="utf-8") as file:
            data = file.read()

        lineno = analyze(data).lloc
        if lineno >= max_entrypoint_count:
            report.add(Record(WARNING,
                              f"Complex entry point. Check: {library} | "
                              f"Counted lines: {lineno} | Lines allowed: {max_entrypoint_count}"))

    except UnicodeDecodeError as e:
        report.add(Record(PROBLEM, f"UnicodeDecodeError: {e}"))

    except SyntaxError as e:
        if e.msg == 'SyntaxError at line: 1':
            report.add(Record(PROBLEM,
                              (f"Error parsing file, is your file saved with UTF-8 encoding? "
                               f"Make sure it has no BOM. Check: {library}")))
        else:
            report.add(Record(PROBLEM,
                               "Error parsing file, is there a syntax error in your file?" \
                               f"Check: {library}"))



"""
    Copyright (C) 2017-2018 Team Kodi
    This file is part of Kodi - kodi.tv

    SPDX-License-Identifier: GPL-3.0-only
    See LICENSES/README.md for more information.
"""

import os
import pathlib
import re

from .common import relative_path
from .record import PROBLEM, Record
from .report import Report


def find_file(name: str, path: str):
    """Looks for a file in a given path.
        :name: name of the file to look for
        :path: path of the directory.
        :return: full path for the file.
    """
    for file_name in os.listdir(path):
        match = re.match(name, file_name, re.IGNORECASE)
        if match is not None:
            return os.path.join(path, match.string)
    return None


def find_files_recursive(name: str, path: str):
    """This looks for a file but only returns the first occurrence
        :name: name of the file to look for
        :path: path of directory to look for the file
    """
    for root, _, files in os.walk(path):
        for file in files:
            if name in file:
                yield os.path.join(root, file)


def create_file_index(path: str):
    """Creates a list having multiple dictionaries in following format:
        [{'name':<file_name>, 'path': '<path_to_file>'}]

        :path: path for the directory
    """
    file_index = []
    for root, folders, files in os.walk(path, topdown=True):
        if ".git" in folders:
            folders.remove(".git")
        for file_name in files:
            file_index.append({"path": root, "name": file_name})
    return file_index


def find_in_file(path: str, search_terms: list, whitelisted_file_types: list):
    """Finds for particular terms in whitelisted file type i.e .py or .xml
        :path: path of a directory
        :search_term: list of all the terms to be searched
        :whitelisted_file_type: list of all the whitelisted file types
    """
    results = []
    if search_terms:
        for directory in os.walk(path):
            for file_name in directory[2]:
                if pathlib.Path(file_name).suffix in whitelisted_file_types or not whitelisted_file_types:
                    file_path = os.path.join(directory[0], file_name)

                    with open(file_path, "r", encoding="utf8") as searchfile:
                        linenumber = 0
                        for line in searchfile:
                            linenumber = linenumber + 1
                            for term in search_terms:
                                if term in line:
                                    results.append({"term": term, "line": line.strip(
                                    ), "searchfile": file_path, "linenumber": linenumber})
    return results


def addon_file_exists(report: Report, addon_path: str, file_name: str):
    """check whether addon file exists or not
        :addon_path: path to the addon
        :file_name: name of the addon file
    """
    if find_file(file_name, addon_path) is None:
        report.add(Record(PROBLEM, f"Not found {file_name} in folder {relative_path(addon_path)}"))



"""
    Copyright (C) 2017-2018 Team Kodi
    This file is part of Kodi - kodi.tv

    SPDX-License-Identifier: GPL-3.0-only
    See LICENSES/README.md for more information.
"""

import os
import re

import polib

from . import handle_files
from .common import relative_path
from .record import INFORMATION, PROBLEM, WARNING, Record
from .report import Report

RE_LANG_CODE = re.compile(r"^[a-z]{2,3}(?:_[a-zA-Z]{2}(?:@\S+)?)?$")


def check_for_legacy_strings_xml(report: Report, addon_path: str):
    """Find for the string.xml file in addon which was used in old versions
        :addon_path: path of the addon
    """
    for file in handle_files.find_files_recursive("strings.xml", os.path.join(addon_path, "resources", "language")):
        report.add(
            Record(PROBLEM, f"Found {relative_path(file)} please migrate to strings.po."))


def find_blacklisted_strings(report: Report, addon_path: str, problems: list, warnings: list, file_types: list):
    """Find for any blacklisted strings in the addons files
        :addon_path: Path of theh addon
        :problems: List of all the strings that will cause problem being in an addon
        :warnings: List of all the strings that shouldn't be in addon
                        but doesn't cause any problem
        :file_type: List of the whitelisted files to look into
    """
    for result in handle_files.find_in_file(addon_path, problems, file_types):
        report.add(Record(PROBLEM, f"Found blacklisted term {result['term']} in file "\
                     f"{result['searchfile']}:{result['linenumber']} ({result['line']})"))

    for result in handle_files.find_in_file(addon_path, warnings, file_types):
        report.add(Record(WARNING, f"Found blacklisted term {result['term']} in file "\
                    f"{result['searchfile']}:{result['linenumber']} ({result['line']})"))


def check_for_invalid_strings_po(report: Report, file_index: list, is_language_addon: bool = False):
    """Validate strings.po files
        :file_index: list having names and path of all the files present in addon
    """
    en_gb_present = False
    report_made = False

    po_file_index = [f for f in file_index if f["name"] == "strings.po"]

    for po_file in po_file_index:
        success, language_code = parse_po_file(report, po_file.get("path"), po_file)
        if not en_gb_present and success and language_code and language_code.lower() == 'en_gb':
            en_gb_present = True

        if not report_made and not success:
            report_made = True

    skip_en_gb_validation = is_language_addon or all(
        (
            _is_using_legacy_language_directory_structure(po_file.get("path")) \
            for po_file in po_file_index
        )
    )

    if not en_gb_present and not skip_en_gb_validation:
        if po_file_index:
            report_made = True
            report.add(Record(PROBLEM, "Required default language 'en_gb' is not present."))
        else:
            report.add(Record(WARNING, "No PO files found. "
                                       "Consider adding language support to your add-on."))

    if po_file_index and not report_made:
        report.add(Record(INFORMATION, "PO files are valid"))


def parse_po_file(report: Report, language_path: str, po_file: dict):
    """Parse strings.po files
        :language_path: base language path, "<ADDON_PATH>/resources/language/resource.language.
        or <ADDON_PATH>/resources/language/language (legacy)"
        :po_file: dict containing name and path of the po file representing a file from file_index
    """
    success = True
    full_path = os.path.join(po_file["path"], po_file["name"])

    language_code = ''
    if len(language_path.rpartition('.')) == 3:
        language_code = language_path.rpartition('.')[2]

    if not _is_using_legacy_language_directory_structure(po_file["path"]) \
        and not RE_LANG_CODE.match(language_code):
        success = False
        report.add(Record(PROBLEM, "PO file with invalid language code in the correct path: " \
            f"{relative_path(full_path)}"))
        return success, language_code

    with open(full_path, "r", encoding="utf-8") as f:
        try:
            contents = f.read()
            f.seek(0)
            lines = f.readlines()
        except UnicodeDecodeError:
            success = False
            report.add(Record(PROBLEM,
                f"Invalid PO file {relative_path(full_path)}: File is not saved with UTF-8 encoding"))
            return success, language_code

    if not contents:
        success = False
        report.add(Record(PROBLEM,
            f"Invalid PO file {relative_path(full_path)}: File is empty"))
        return success, language_code

    if "\r\n" in contents:
        success = False
        report.add(Record(WARNING,
            f"Windows line endings found in {relative_path(full_path)}, consider converting to Linux line endings."
        ))

    header = contents[:contents.find("msgctxt \"#")]
    if not re.search(r'msgid ""\s+msgstr ""', header):
        # This is only required by polib if metadata follows, Kodi requires this regardless of metadata
        success = False
        report.add(Record(PROBLEM, f"Invalid PO file {relative_path(full_path)}:\nMissing required header:\n"
                                   "\tmsgid \"\"\n\tmsgstr \"\""))

    if contents[0] == "\ufeff":
        success = False
        report.add(Record(PROBLEM, f"Invalid PO file {relative_path(full_path)}: File contains BOM (byte order mark)"))
        return success, language_code

    # fix any Gettext automatic comments in the source before testing with polib
    contents = ''.join(['# ' + line.replace('#', '').lstrip()
                        if (line.startswith('#') and line.replace('#', '').lstrip())
                        else ('\n'
                              if line.startswith('#') and not line.replace('#', '').lstrip()
                              else line)
                        for line in lines])
    try:
        polib.pofile(contents, encoding="utf-8")
    except OSError as error:
        # raised on the first syntax error
        message = str(error)
        patterns = [r"\(line (?P<line_num>[0-9]+)\)\s*:\s*(?P<message>[\s\S]+)\s*$",
                    r"(?P<message>Syntax error)[\s\S]+\(line (?P<line_num>[0-9]+)\)\s*$"]
        for pattern in patterns:
            match = re.search(pattern, message)
            if match:
                # restructure message to remove file and path
                message = f"{match.group('message')} on line {match.group('line_num')}"
                break

        success = False
        report.add(Record(PROBLEM, f"Invalid PO file {relative_path(full_path)}: {message}"))

    return success, language_code


def _is_using_legacy_language_directory_structure(po_file: str):
    """Checks if all the po files in the addon directory are not
    using the new language folder layout

    Args:
        po_file (str): Path to the po file

    Returns:
        [bool]: If addon is using legacy directory path
    """

    return "resource.language." not in po_file
